#!/usr/bin/python
# -*- coding=utf-8 -*-
"""
Arabic Transliteration routins
@author: Taha Zerrouki
@contact: taha dot zerrouki at gmail dot com
@license: GPL
@date:2018/08/146
@version: 0.1
"""
import re

from . import araby as ar

t2a_table = {
    "A": ar.ALEF,
    "b": ar.BEH,
    "t": ar.TEH,
    "p": ar.TEH_MARBUTA,
    "v": ar.THEH,
    "j": ar.JEEM,
    "H": ar.HAH,
    "x": ar.KHAH,
    "d": ar.DAL,
    "*": ar.THAL,
    "r": ar.REH,
    "z": ar.ZAIN,
    "s": ar.SEEN,
    "$": ar.SHEEN,
    "S": ar.SAD,
    "D": ar.DAD,
    "T": ar.TAH,
    "Z": ar.ZAH,
    "E": ar.AIN,
    "g": ar.GHAIN,
    "f": ar.FEH,
    "q": ar.QAF,
    "k": ar.KAF,
    "l": ar.LAM,
    "m": ar.MEEM,
    "n": ar.NOON,
    "h": ar.HEH,
    "w": ar.WAW,
    "y": ar.YEH,
    "Y": ar.ALEF_MAKSURA,
    "'": ar.HAMZA,
    "&": ar.WAW_HAMZA,
    ">": ar.ALEF_HAMZA_ABOVE,
    "<": ar.ALEF_HAMZA_BELOW,
    "|": ar.ALEF_MADDA,
    "}": ar.YEH_HAMZA,
    "_": ar.TATWEEL,
    "a": ar.FATHA,
    "F": ar.FATHATAN,
    "i": ar.KASRA,
    "K": ar.KASRATAN,
    "u": ar.DAMMA,
    "N": ar.DAMMATAN,
    "~": ar.SHADDA,
    "o": ar.SUKUN,
    "`": ar.MINI_ALEF,
    # ~ '{': ar.ALEF_WASLA
    "{": ar.ALEF,
}


# conversion Tablke from the tim bulwalter represetation
# into sampa notation
t2sampa_table = {
    "A": "a:",  # ALEF,
    "b": "b",  # BEH
    "t": "t",  # TEH
    "p": "h",  # TEH_MARBUTA,
    "v": "T",  # THEH
    "j": "g",  # JEEM,
    "H": "x",  # HAH,
    "x": "X",  # KHAH,
    "d": "d",  # DAL,
    "*": "D",  # THAL,
    "r": "r",  # REH,
    "z": "z",  # ZAIN,
    "s": "s",  # SEEN,
    "$": "S",  # SHEEN,
    "S": "s'",  # SAD,
    "D": "d'",  # DAD,
    "T": "t'",  # TAH,
    "Z": "D'",  # ZAH,
    "E": "?'",  # AIN,
    "g": "G",  # GHAIN,
    "f": "f",  # FEH,
    "q": "q",  # QAF,
    "k": "k",  # KAF,
    "l": "l",  # LAM,
    "m": "m",  # MEEM,
    "n": "n",  # NOON,
    "h": "h",  # HEH,
    "w": "w",  # WAW,
    "y": "j",  # YEH,
    "Y": ":",  # ALEF_MAKSURA,
    "'": "?",  # HAMZA,
    "&": "?",  # WAW_HAMZA,
    ">": "?",  # ALEF_HAMZA_ABOVE,
    "<": "?",  # ALEF_HAMZA_BELOW,
    "|": "?a:",  # ALEF_MADDA,
    "}": "?",  # YEH_HAMZA,
    "_": "",  # '',#TATWEEL,
    "a": "a",  # FATHA,
    "F": "an",  # FATHATAN,
    "i": "i",  # KASRA,
    "K": "in",  # KASRATAN,
    "u": "u",  # DAMMA,
    "N": "un",  # DAMMATAN,
    "~": ar.SHADDA,
    "o": "",  # SUKUN,
    "`": "a:",  # MINI_ALEF,
    "{": "",  # ALEF_WASLA,
}


a2en_table = {
    "ء": "2",
    "آ": "A",
    "أ": "A",
    "ؤ": "2",
    "إ": "2",
    "ئ": "2",
    "ا": "A",
    "ب": "b",
    "ة": "t",
    "ت": "t",
    "ث": "th",
    "ج": "j",
    "ح": "H",
    "خ": "kh",
    "د": "d",
    "ذ": "dh",
    "ر": "r",
    "ز": "z",
    "س": "s",
    "ش": "sh",
    "ص": "S",
    "ض": "D",
    "ط": "T",
    "ظ": "zh",
    "ع": "E",
    "غ": "g",
    "ـ": "",
    "ف": "f",
    "ق": "q",
    "ك": "k",
    "ل": "l",
    "م": "m",
    "ن": "n",
    "ه": "h",
    "و": "w",
    "ى": "a",
    "ي": "y",
    "ً": "an",
    "ٌ": "un",
    "ٍ": "in",
    "َ": "a",
    "ُ": "u",
    "ِ": "i",
    "ّ": "",
    "ْ": "",
    "ٰ": "a",
}

a2t_table = {v: k for k, v in t2a_table.items()}
# correct case
a2t_table[ar.ALEF] = "A"

T2D_TRANS = str.maketrans(ar.NOT_DEF_HARAKA + ar.TASHKEEL_STRING, "012345678")
T2A_TRANS = str.maketrans(ar.NOT_DEF_HARAKA + ar.TASHKEEL_STRING, "0AUIauio3")
D2T_TRANS = str.maketrans("012345678", ar.NOT_DEF_HARAKA + ar.TASHKEEL_STRING)
A2T_TRANS = str.maketrans("0AUIauio3", ar.NOT_DEF_HARAKA + ar.TASHKEEL_STRING)
# -------------- Digits Trans -------------- #
E2W_TRANS = str.maketrans("".join(ar.NUMBERS_EAST), "".join(ar.NUMBERS_WEST))
E2P_TRANS = str.maketrans("".join(ar.NUMBERS_EAST), "".join(ar.NUMBERS_PERS))
W2E_TRANS = str.maketrans("".join(ar.NUMBERS_WEST), "".join(ar.NUMBERS_EAST))
W2P_TRANS = str.maketrans("".join(ar.NUMBERS_WEST), "".join(ar.NUMBERS_PERS))
P2E_TRANS = str.maketrans("".join(ar.NUMBERS_PERS), "".join(ar.NUMBERS_EAST))
P2W_TRANS = str.maketrans("".join(ar.NUMBERS_PERS), "".join(ar.NUMBERS_WEST))


def translate(word, table):
    """translate a word accoring to table"""
    return word.translate(table)


def tim2utf8(s):
    "Tranliteration to UTF-8 conversion of a string"
    mystr = ""
    for mychar in s:
        mystr += t2a_table.get(mychar, mychar)
    return mystr


def utf82tim(s):
    "Tranliteration to Tim Buckwalter conversion of a string"
    mystr = ""
    for mychar in s:
        mystr += a2t_table.get(mychar, mychar)
    return mystr


def convertShadda(word, shadda="~"):
    if word[0] == shadda:
        # to avoid that shadda is present in the begining
        word = word[1:]
    while shadda in word:
        i = word.index(shadda)
        if i - 1 >= 0:
            # replace the letter before shadda to double
            # replace the first one only
            word = word.replace(shadda, word[i - 1], 1)
    return word


def tim2sampa(s):
    """Tranliteration to SAMPA code phonemes conversion of a string
    We suppose that all words are full vocalized.
    We convert according to t2s table,
    and the shadda is converted to double letter
    """
    mystr = ""
    # convert shadda
    # the conversion is made before translatiration,
    # to avoid errors on phonems which have two letters as s[ or gH
    s = convertShadda(s)
    for mychar in s:
        mystr += t2sampa_table.get(mychar, mychar)

    # convert waw and yeh after damma and kasra
    mystr = re.sub("(?<=u)w", ":", mystr)
    mystr = re.sub("(?<=i)j", ":", mystr)
    return mystr


def utf82latin(s):
    "Tranliteration from UTF-8  to latin with plain english no symbol"
    mystr = ""
    for mychar in s:
        mystr += a2en_table.get(mychar, mychar)
    return mystr


def convert(text, code_from, code_to):
    """
    convert text from code_from to code_to

    """
    code1 = code_from.lower()
    code2 = code_to.lower()
    if code1 in ("utf", "utf8", "arabic"):
        if code2 in ("tim", "buckwalter"):
            return utf82tim(text)
        elif code2 == "sampa":
            return tim2sampa(utf82tim(text))
        elif code2 in ("latin", "ascii"):
            return utf82latin(text)
        else:
            return text

    if code1 in ("tim", "buckwalter"):
        if code2 in ("utf", "utf8", "arabic"):
            return tim2utf8(text)
        elif code2 == "sampa":
            return tim2sampa(text)
        else:
            return text


def segment_language(text):
    """
    Detect language
    """
    if not text:
        return text
    resultlist = []
    if re.search("[\u0600-\u06ff]", text[0]):
        # ~ if re.search(u"[\u0600-\u06ff]", text):
        arabic = True
    else:
        arabic = False
    actual_text = ""
    for k in text:
        if re.search("[\u0600-\u06ff]", k):
            if arabic:
                actual_text += k
            else:
                resultlist.append(("latin", actual_text))
                arabic = True
                actual_text = k
        elif re.search("[\s\d\?, :\!\(\)]", k):
            actual_text += k
        else:
            if arabic:
                i = len(actual_text)
                temp_text = ""
                while not re.search("[\u0600-\u06ff]", actual_text[i : i + 1]):
                    i -= 1
                temp_text = actual_text[i + 1 :]
                actual_text = actual_text[: i + 1]
                resultlist.append(("arabic", actual_text))
                arabic = False
                actual_text = temp_text + k
            else:
                actual_text += k
    if arabic:
        resultlist.append(("arabic", actual_text))
    else:
        resultlist.append(("latin", actual_text))
    return resultlist


def delimite_language(text, language="arabic", start="<", end=">"):
    new_chunks_list = []
    chunks = segment_language(text)
    for lang, chunk in chunks:
        if lang == language:
            new_chunks_list.append("%s%s%s" % (start, chunk, end))
        else:
            new_chunks_list.append(chunk)
    return " ".join(new_chunks_list)


def normalize_digits(text, source="all", out="west"):
    """
    Normalize digits to and from the following writing systems:
    west:    Western Arabic numerals                (0123456789)
    east:    Eastern Arabic (Hindu-Arabic) numerals (٠١٢٣٤٥٦٧٨٩)
    persian: Persian/Urdu numerals                  (۰۱۲۳۴۵۶۷۸۹)

    if `source = all`, then all digits contained in the text
    will be normalized into `out` writing system.
    Otherwise digits written in `source` will be normalized
    without affecting the rest of the digits.

    Example:
        >>> text = u'۰۱۲۳۴۵۶۷۸۹ ٠١٢٣٤٥٦٧٨٩ 123456789'
        >>> normalize_digits(text, source='all', out='west')
        '0123456789 0123456789 0123456789'
        >>> normalize_digits(text, source='persian', out='west')
        >>> '0123456789 ٠١٢٣٤٥٦٧٨٩ 0123456789'

    @param text: unnormalized text.
    @type text: unicode.
    @param source: Writing system for the digits to be normalized.
                   (default is all).
    @type source: string
    @param out: Intended writing system for source.
                (default is west)
    @return: returns a normalized text.
    @rtype: unicode.
    """
    source = source.lower()
    out = out.lower()
    assert source in ["all", "west", "east", "persian"], (
        "Invalid option for `source`: %s" % source
    )
    assert out in ["west", "east", "persian"], "Invalid option for `out`: %s" % out
    if source == out:
        return text
    source_to_out_tbl = {
        "west": {"east": W2E_TRANS, "persian": W2P_TRANS},
        "east": {"west": E2W_TRANS, "persian": E2P_TRANS},
        "persian": {"west": P2W_TRANS, "east": P2E_TRANS},
    }
    if source == "all":
        del source_to_out_tbl[out]
        for tbl in source_to_out_tbl.values():
            text = translate(text, tbl[out])
        return text
    return translate(text, source_to_out_tbl[source][out])


def encode_tashkeel(word, method="ascii"):
    """
    Encode word marks into decimal or Ascii string to be saved as integer

    Example:
        >>> import pyarabic.trans
        >>> word1 = u"هَارِبًا"
        >>> pyarabic.trans.encode_tashkeel(word1)
        ('هاربا', 'a0iA0')
        >>> pyarabic.trans.encode_tashkeel(word1, "decimal")
        ('هاربا', 40610)
        >>> letters = u"هاربا"
        >>> encoded_marks = u"a0iA0"
        >>> pyarabic.trans.decode_tashkeel(letters, encoded_marks)
        'هَارِبًا'
        >>> letters = u"هاربا"
        >>> encoded_marks = 40610
        >>> pyarabic.trans.decode_tashkeel(letters, encoded_marks, "decimal")
        'هَارِبًا'


    @input word: diacritized arabic diacritcs
    @type word: unicode
    @return:  (letters, encoded) zero if fails
    @rtype: (letters, encoded) ttring/ integer
    """
    letters, marks = ar.separate(word)

    if method == "decimal":
        transed = translate(marks, T2D_TRANS)
    elif method == "ascii":
        transed = translate(marks, T2A_TRANS)
    else:
        transed = translate(marks, T2A_TRANS)

    if method == "decimal":
        try:
            transed = int(transed)
        except:
            return word, ""
    return letters, transed


def decode_tashkeel(word, marks, method="ascii"):
    """decode tashkeel"""
    """
    decode marks from decimal/ascii string to be joint on word
    @input word: undiacritized arabic diacritcs
    @type word: unicode
    @input marks: encoded marks
    @type marks: unicode/integer
    @return:  diacritized word
    @rtype: unicode
    """
    if type(marks) != (str):
        marks = str(marks)
    # zeros can be removed in int code, then we must add them to left
    marks = marks.rjust(len(word), str("0"))
    if method == "decimal":
        transed = translate(marks, D2T_TRANS)
    elif method == "ascii":
        transed = translate(marks, A2T_TRANS)
    else:
        transed = translate(marks, A2T_TRANS)
    word2 = ar.joint(word, transed)
    return word2


if __name__ == "__main__":

    words = """qulo
>aEuw*u
bi
rab~i
{l
n~aAsi
maliki
{l
n~aAsi
<ila`hi
{l
n~aAsi
min
$ar~i
{lo
wasowaAsi
{lo
xan~aAsi
{l~a*iY
yuwasowisu
fiY
Suduwri
{l
n~aAsi
mina
{lo
jin~api
wa
{l
n~aAsi""".split(
        "\n"
    )
    for word in words:
        arabic = tim2utf8(word)
        timu = utf82tim(arabic)
        sampa = tim2sampa(word)
        arabic2 = convert(word, "tim", "utf")
        timu2 = convert(arabic2, "utf", "tim")  # utf82tim(arabic)
        sampa2 = convert(word, "tim", "sampa")  # tim2sampa(word)
        print(
            "\t".join(
                [
                    word,
                    arabic,
                    timu,
                    sampa,
                    arabic2,
                    timu2,
                    sampa2,
                    str(timu == word),
                    str(arabic2 == arabic),
                    str(timu2 == timu),
                    str(sampa2 == sampa),
                ]
            ).encode("utf8")
        )
        # ~ print(u'\t'.join([word, tim2sampa(word)]).encode('utf8'))

    utf2tim_table = {v: k for k, v in t2a_table.items()}
    print(utf2tim_table)

    from arabrepr import arepr

    # test detect language
    text = """السلام عليكم how are you, لم اسمع أخبارك منذ مدة, where are you going"""
    print(arepr(segment_language(text)))
    text_out = delimite_language(text, start="\RL{", end="}")
    print(text_out.encode("utf8"))
    text_out = delimite_language(text, start="<arabic>", end="</arabic>")
    print(text_out.encode("utf8"))



import json
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Mapping, Optional, Sequence
from phoonnx.util import LOG
from phoonnx.phoneme_ids import (load_phoneme_ids, BlankBetween,
                                 DEFAULT_BLANK_WORD_TOKEN, DEFAULT_BLANK_TOKEN,
                                 DEFAULT_PAD_TOKEN, DEFAULT_BOS_TOKEN, DEFAULT_EOS_TOKEN)

DEFAULT_NOISE_SCALE = 0.667
DEFAULT_LENGTH_SCALE = 1.0
DEFAULT_NOISE_W_SCALE = 0.8


class Engine(str, Enum):
    """voices trained with these frameworks are explicitly supported.
    This mainly affects the format of .json file and possibly tokenization"""
    PHOONNX = "phoonnx"
    PIPER = "piper"
    MIMIC3 = "mimic3"
    COQUI = "coqui"


class Alphabet(str, Enum):
    UNICODE = "unicode"
    IPA = "ipa"
    ARPA = "arpa" # en
    SAMPA = "sampa"
    XSAMPA = "x-sampa"
    HANGUL = "hangul" # ko
    KANA = "kana" # ja
    HIRA = "hira" # ja
    HEPBURN = "hepburn" # ja romanization
    KUNREI = "kunrei" # ja romanization
    NIHON = "nihon" # ja romanization
    PINYIN = "pinyin" # zh
    ERAAB = "eraab" # fa
    COTOVIA = "cotovia" # gl
    HANZI = "hanzi" # zh
    BUCKWALTER = "buckwalter" # ar



class PhonemeType(str, Enum):
    RAW = "raw"  # direct phonemes
    UNICODE = "unicode"  # unicode codepoints
    GRAPHEMES = "graphemes" # text characters

    MISAKI = "misaki"
    ESPEAK = "espeak"
    GRUUT = "gruut"
    GORUUT = "goruut"
    EPITRAN = "epitran"
    BYT5 = "byt5"
    CHARSIU = "charsiu"  # technically same as byt5, but needs special handling for whitespace
    TRANSPHONE = "transphone"
    MIRANDESE = "mwl_phonemizer"

    DEEPPHONEMIZER = "deepphonemizer" # en
    OPENPHONEMIZER = "openphonemizer" # en
    G2PEN = "g2pen" # en

    TUGAPHONE = "tugaphone"  # pt
    G2PFA = "g2pfa"
    OPENJTALK = "openjtalk" # ja
    CUTLET = "cutlet" # ja
    PYKAKASI = "pykakasi" # ja
    COTOVIA = "cotovia"  # galician  (no ipa!)
    PHONIKUD = "phonikud"  # hebrew
    MANTOQ = "mantoq"  # arabic
    VIPHONEME = "viphoneme" # vietnamese
    G2PK = "g2pk" # korean
    KOG2PK = "kog2p" # korean
    G2PC = "g2pc" # chinese
    G2PM = "g2pm" # chinese
    PYPINYIN = "pypinyin" # chinese
    XPINYIN = "xpinyin" # chinese
    JIEBA = "jieba" # chinese  (not a real phonemizer!)


@dataclass
class VoiceConfig:
    """TTS model configuration"""

    num_symbols: int
    """Number of phonemes."""

    num_speakers: int
    """Number of speakers."""

    num_langs: int
    """Number of langs."""

    sample_rate: int
    """Sample rate of output audio."""

    lang_code: Optional[str]
    """Name of espeak-ng voice or alphabet."""

    phoneme_id_map: Optional[Mapping[str, Sequence[int]]]
    """Phoneme -> [id,]. Used for phoneme-based models."""

    phoneme_type: PhonemeType
    """espeak, byt5, text, cotovia, or graphemes."""

    alphabet: Optional[Alphabet]

    phonemizer_model: Optional[str]
    """for phonemizers that allow changing base model """

    speaker_id_map: Mapping[str, int] = field(default_factory=dict)
    """Speaker -> id"""

    lang_id_map: Mapping[str, int] = field(default_factory=dict)
    """lang-code -> id"""

    # Info about what framework was used to train the model
    engine: Engine = Engine.PHOONNX

    # Inference settings
    length_scale: float = DEFAULT_LENGTH_SCALE
    noise_scale: float = DEFAULT_NOISE_SCALE
    noise_w_scale: float = DEFAULT_NOISE_W_SCALE
    add_diacritics: bool = None # arabic and hebrew

    # tokenization settings
    blank_at_start: bool = True
    blank_at_end: bool = True
    include_whitespace: Optional[bool] = True
    pad_token: Optional[str] = DEFAULT_PAD_TOKEN
    blank_token: Optional[str] = DEFAULT_PAD_TOKEN
    bos_token: Optional[str] = DEFAULT_BOS_TOKEN
    eos_token: Optional[str] = DEFAULT_EOS_TOKEN
    word_sep_token: Optional[str] = DEFAULT_BLANK_WORD_TOKEN
    blank_between: BlankBetween = BlankBetween.TOKENS_AND_WORDS

    def __post_init__(self):
        """
        Finalize dataclass defaults after initialization.
        
        If `add_diacritics` is None, sets it to False; if `lang_code` is present and starts with "ar", sets `add_diacritics` to True. Ensures `lang_code` is set to "und" when not provided.
        """
        if self.add_diacritics is None:
            self.add_diacritics = False
            if self.lang_code and self.lang_code.startswith("ar"):
                self.add_diacritics = True
        self.lang_code = self.lang_code or "und"

    @staticmethod
    def is_mimic3(config: dict[str, Any]) -> bool:
        # https://huggingface.co/mukowaty/mimic3-voices

        # mimic3 models indicate a phonemizer strategy in their config
        if ("phonemizer" not in config or
                not isinstance(config["phonemizer"], str)):
            return False

        # mimic3 models include a "phonemes" section with token info
        if "phonemes" not in config or not isinstance(config["phonemes"], dict):
            return False

        # validate phonemizer type as expected by mimic3
        phonemizer = config["phonemizer"]
        # class Phonemizer(str, Enum):
        #     SYMBOLS = "symbols"
        #     GRUUT = "gruut"
        #     ESPEAK = "espeak"
        #     EPITRAN = "epitran"
        if phonemizer not in ["symbols", "gruut", "espeak", "epitran"]:
            return False

        return True

    @staticmethod
    def is_piper(config: dict[str, Any]) -> bool:
        if "piper_version" in config:
            return True
        # piper models indicate a phonemizer strategy in their config
        if ("phoneme_type" not in config or
                not isinstance(config["phoneme_type"], str)):
            return False

        # piper models include a "phoneme_id_map" section mapping phonemes to int
        if "phoneme_id_map" not in config or not isinstance(config["phoneme_id_map"], dict):
            return False

        # validate phonemizer type as expected by piper
        phonemizer = config["phoneme_type"]
        if phonemizer not in ["text", "espeak"]:
            return False

        return True

    @staticmethod
    def is_coqui_vits(config: dict[str, Any]) -> bool:
        # coqui vits grapheme models include a "characters" section with token info
        if "characters" not in config or not isinstance(config["characters"], dict):
            return False

        # double check this was trained with coqui
        if config["characters"].get("characters_class", "") not in ["TTS.tts.models.vits.VitsCharacters",
                                                                    "TTS.tts.utils.text.characters.Graphemes"]:
            return False

        return True

    @staticmethod
    def is_phoonnx(config: dict[str, Any]) -> bool:
        return "phoonnx_version" in config

    @staticmethod
    def from_dict(config: dict[str, Any],
                  phonemes_txt: Optional[str] = None,
                  lang_code: Optional[str] = None,
                  phoneme_type_str: Optional[str] = None) -> "VoiceConfig":
        """
        Create a VoiceConfig from a configuration dictionary and an optional external phonemes file.
        
        Constructs a VoiceConfig by deriving engine, alphabet, phoneme mapping, and inference/synthesis settings from the provided config and optional phonemes_txt. Detects model type (Phoonnx, Piper, Mimic3, Coqui) and applies model-specific defaults and mappings; optional lang_code and phoneme_type_str override values in the config.
        
        Parameters:
            config (dict[str, Any]): Parsed model configuration dictionary.
            phonemes_txt (Optional[str]): Path to an external phonemes file (.txt or .json) used to build or override the phoneme id mapping.
            lang_code (Optional[str]): Optional language code to override the config's language.
            phoneme_type_str (Optional[str]): Optional phoneme type name to override the config's phoneme type.
        
        Returns:
            VoiceConfig: A populated VoiceConfig instance with fields set from the config and any provided phonemes file.
        
        Raises:
            ValueError: If the model is detected as Mimic3 but no phonemes_txt is provided.
        """
        blank_type = BlankBetween.TOKENS_AND_WORDS
        lang_code = lang_code or config.get("lang_code")
        phoneme_type_str = phoneme_type_str or config.get("phoneme_type")
        phoneme_id_map = config.get("phoneme_id_map")
        alphabet = config.get("alphabet")
        engine = Engine.PHOONNX
        diacritics = False

        if phonemes_txt:
            if phonemes_txt.endswith(".txt"):
                # either from mimic3 models or as an override at runtime
                with open(phonemes_txt, "r", encoding="utf-8") as ids_file:
                    phoneme_id_map = load_phoneme_ids(ids_file)
            elif phonemes_txt.endswith(".json"):
                with open(phonemes_txt) as ids_file:
                    phoneme_id_map = json.load(ids_file)

        if VoiceConfig.is_phoonnx(config):
            engine = Engine.PHOONNX

            lang_code = lang_code or config.get("lang_code")
            phoneme_type_str = config.get("phoneme_type", PhonemeType.ESPEAK.value)
            alphabet = Alphabet(config.get("alphabet", "ipa"))
            diacritics = config.get("inference", {}).get("add_diacritics", True)

            config["pad"] =  DEFAULT_PAD_TOKEN
            config["blank"] = DEFAULT_BLANK_TOKEN
            config["bos"] = DEFAULT_BOS_TOKEN
            config["eos"] = DEFAULT_EOS_TOKEN
        # check if model was trained for PiperTTS
        elif VoiceConfig.is_piper(config):
            engine = Engine.PIPER

            lang_code = lang_code or (config.get("language", {}).get("code") or
                         config.get("espeak", {}).get("voice"))
            diacritics = lang_code.startswith("ar")
            phoneme_type_str = config.get("phoneme_type", PhonemeType.ESPEAK.value)
            if phoneme_type_str == "text":
                phoneme_type_str = PhonemeType.UNICODE.value
                alphabet = Alphabet.UNICODE
            else:
                alphabet = Alphabet.IPA

            # not configurable in piper
            config["pad"] =  DEFAULT_PAD_TOKEN
            config["blank"] = DEFAULT_BLANK_TOKEN
            config["bos"] = DEFAULT_BOS_TOKEN
            config["eos"] = DEFAULT_EOS_TOKEN

        # check if model was trained for Mimic3
        elif VoiceConfig.is_mimic3(config):
            engine = Engine.MIMIC3

            if not phonemes_txt:
                raise ValueError("mimic3 models require an external phonemes.txt file in addition to the config")
            lang_code = config.get("text_language")
            phoneme_type_str = config.get("phonemizer", PhonemeType.GRUUT.value)
            # read phoneme settings
            phoneme_cfg = config.get("phonemes", {})
            blank_type = BlankBetween(phoneme_cfg.get("blank_between", "tokens_and_words"))
            config.update(phoneme_cfg)

            if phoneme_type_str == "symbols":
                # Mimic3 "symbols" models are grapheme models
                # symbol map comes from phonemes_txt
                phoneme_type_str = PhonemeType.GRAPHEMES.value
                alphabet = Alphabet.UNICODE
            else:
                alphabet = Alphabet.IPA

        # check if model was trained with Coqui
        elif VoiceConfig.is_coqui_vits(config):
            engine = Engine.COQUI
            phoneme_type_str = PhonemeType.GRAPHEMES.value
            alphabet = Alphabet.UNICODE

            # NOTE: lang code usually not provided and often wrong :(
            ds = config.get("datasets", [])
            if ds and not lang_code:
                lang_code = ds[0].get("language")

            characters_config = config.get("characters", {})
            if config.get("add_blank", True):
                blank_type = BlankBetween.TOKENS
                characters_config["blank"] = characters_config.get("blank") or "<BLNK>"
            config.update(characters_config)
            # For Coqui VITS grapheme models, build phoneme_id_map from characters
            characters = characters_config.get("characters")
            punctuations = characters_config.get("punctuations")

            if not config.get("enable_eos_bos_chars", True):
                config["bos"] = config["eos"] = None

            # Construct vocabulary based on the order defined in the original Graphemes class
            # [PAD, EOS, BOS, BLANK, CHARACTERS, PUNCTUATIONS]
            vocab_list = []

            if characters_config.get("pad") is not None:
                vocab_list.append(characters_config["pad"])

            # ?? - haven't see any coqui model
            # adding bos and eos to vocab_list

            #if characters_config.get("eos") is not None:
            #    vocab_list.append(characters_config["eos"])
            #if characters_config.get("bos") is not None:
            #    vocab_list.append(characters_config["bos"])

            if punctuations:
                vocab_list.extend(list(punctuations))
            if characters:
                vocab_list.extend(list(characters))


            if characters_config.get("blank") is not None:
                vocab_list.append(characters_config["blank"])

            # Ensure unique characters and sort if needed (though not strictly necessary for map creation)
            # This part of logic was previously in Graphemes, now implicitly handled by set/list conversion
            phoneme_id_map = {char: idx for idx, char in enumerate(vocab_list)}

        phoneme_type = PhonemeType(phoneme_type_str)
        LOG.debug(f"phonemizer: {phoneme_type}")
        inference = config.get("inference", {})

        include_whitespace = " " in config.get("characters", "") or " " in config.get("phoneme_id_map", {})
        return VoiceConfig(
            num_langs=config.get("num_langs", 1),
            num_symbols=config.get("num_symbols", 256),
            num_speakers=config.get("num_speakers", 1),
            sample_rate=config.get("audio", {}).get("sample_rate", 16000),
            noise_scale=inference.get("noise_scale", DEFAULT_NOISE_SCALE),
            length_scale=inference.get("length_scale", DEFAULT_LENGTH_SCALE),
            noise_w_scale=inference.get("noise_w", DEFAULT_NOISE_W_SCALE),
            add_diacritics=diacritics,
            lang_code=lang_code,
            alphabet=alphabet,
            engine=engine,
            phonemizer_model=config.get("phonemizer_model"),
            phoneme_id_map=phoneme_id_map,
            phoneme_type=phoneme_type,
            speaker_id_map=config.get("speaker_id_map", {}),
            blank_between=blank_type,
            include_whitespace=include_whitespace,
            blank_at_start=config.get("blank_at_start", True),
            blank_at_end=config.get("blank_at_end", True),
            pad_token=config.get("pad"),
            blank_token=config.get("blank"),
            bos_token=config.get("bos"),
            eos_token=config.get("eos"),
            word_sep_token=config.get("word_sep_token") or config.get("blank_word", " ")
        )


@dataclass
class SynthesisConfig:
    """Configuration for synthesis."""

    speaker_id: Optional[int] = None
    """Index of speaker to use (multi-speaker voices only)."""

    lang_id: Optional[int] = None
    """Index of lang to use (multi-lang voices only)."""

    length_scale: Optional[float] = None
    """Phoneme length scale (< 1 is faster, > 1 is slower)."""

    noise_scale: Optional[float] = None
    """Amount of generator noise to add."""

    noise_w_scale: Optional[float] = None
    """Amount of phoneme width noise to add."""

    normalize_audio: bool = True
    """Enable/disable scaling audio samples to fit full range."""

    volume: float = 1.0
    """Multiplier for audio samples (< 1 is quieter, > 1 is louder)."""

    enable_phonetic_spellings: bool = True

    """for arabic and hebrew models"""
    add_diacritics: bool = True


def get_phonemizer(phoneme_type: PhonemeType,
                   alphabet: Alphabet = Alphabet.IPA,
                   model: Optional[str] = None) -> 'Phonemizer':
    from phoonnx.phonemizers import (EpitranPhonemizer, EspeakPhonemizer, OpenPhonemizer, OpenJTaklPhonemizer,
                                     ByT5Phonemizer, CharsiuPhonemizer, DeepPhonemizer, PersianPhonemizer,
                                     G2pCPhonemizer, G2pMPhonemizer, G2PKPhonemizer, G2PEnPhonemizer,
                                     TransphonePhonemizer, MirandesePhonemizer, GoruutPhonemizer, TugaphonePhonemizer,
                                     GruutPhonemizer, GraphemePhonemizer, MantoqPhonemizer, MisakiPhonemizer,
                                     KoG2PPhonemizer, PypinyinPhonemizer, PyKakasiPhonemizer, CotoviaPhonemizer,
                                     CutletPhonemizer, PhonikudPhonemizer, VIPhonemePhonemizer, XpinyinPhonemizer,
                                     UnicodeCodepointPhonemizer, JiebaPhonemizer, RawPhonemes)
    if phoneme_type == PhonemeType.ESPEAK:
        phonemizer = EspeakPhonemizer()
    elif phoneme_type == PhonemeType.BYT5:
        phonemizer = ByT5Phonemizer(model)
    elif phoneme_type == PhonemeType.TUGAPHONE:
        phonemizer = TugaphonePhonemizer()
    elif phoneme_type == PhonemeType.CHARSIU:
        phonemizer = CharsiuPhonemizer(model)
    elif phoneme_type == PhonemeType.GRUUT:
        phonemizer = GruutPhonemizer()
    elif phoneme_type == PhonemeType.GORUUT:
        phonemizer = GoruutPhonemizer()
    elif phoneme_type == PhonemeType.EPITRAN:
        phonemizer = EpitranPhonemizer()
    elif phoneme_type == PhonemeType.MISAKI:
        phonemizer = MisakiPhonemizer()
    elif phoneme_type == PhonemeType.TRANSPHONE:
        phonemizer = TransphonePhonemizer()
    elif phoneme_type == PhonemeType.MIRANDESE:
        phonemizer = MirandesePhonemizer()
    elif phoneme_type == PhonemeType.DEEPPHONEMIZER:
        phonemizer = DeepPhonemizer(model)
    elif phoneme_type == PhonemeType.OPENPHONEMIZER:
        phonemizer = OpenPhonemizer()
    elif phoneme_type == PhonemeType.G2PEN:
        phonemizer = G2PEnPhonemizer(alphabet=alphabet)
    elif phoneme_type == PhonemeType.OPENJTALK:
        phonemizer = OpenJTaklPhonemizer(alphabet=alphabet)
    elif phoneme_type == PhonemeType.PYKAKASI:
        phonemizer = PyKakasiPhonemizer(alphabet=alphabet)
    elif phoneme_type == PhonemeType.CUTLET:
        phonemizer = CutletPhonemizer(alphabet=alphabet)
    elif phoneme_type == PhonemeType.G2PFA:
        phonemizer = PersianPhonemizer(alphabet=alphabet)
    elif phoneme_type == PhonemeType.PHONIKUD:
        phonemizer = PhonikudPhonemizer()
    elif phoneme_type == PhonemeType.MANTOQ:
        phonemizer = MantoqPhonemizer()
    elif phoneme_type == PhonemeType.VIPHONEME:
        phonemizer = VIPhonemePhonemizer()
    elif phoneme_type == PhonemeType.KOG2PK:
        phonemizer = KoG2PPhonemizer(alphabet=alphabet)
    elif phoneme_type == PhonemeType.G2PK:
        phonemizer = G2PKPhonemizer(alphabet=alphabet)
    elif phoneme_type == PhonemeType.PYPINYIN:
        phonemizer = PypinyinPhonemizer(alphabet=alphabet)
    elif phoneme_type == PhonemeType.XPINYIN:
        phonemizer = XpinyinPhonemizer(alphabet=alphabet)
    elif phoneme_type == PhonemeType.JIEBA:
        phonemizer = JiebaPhonemizer()
    elif phoneme_type == PhonemeType.G2PC:
        phonemizer = G2pCPhonemizer(alphabet=alphabet)
    elif phoneme_type == PhonemeType.G2PM:
        phonemizer = G2pMPhonemizer(alphabet=alphabet)
    elif phoneme_type == PhonemeType.COTOVIA:
        phonemizer = CotoviaPhonemizer()
    elif phoneme_type == PhonemeType.UNICODE:
        phonemizer = UnicodeCodepointPhonemizer()
    elif phoneme_type == PhonemeType.GRAPHEMES:
        phonemizer = GraphemePhonemizer()
    elif phoneme_type == PhonemeType.RAW:
        phonemizer = RawPhonemes()
    else:
        raise ValueError("invalid phonemizer")
    return phonemizer



if __name__ == "__main__":
    config_files = [
        "/home/miro/PycharmProjects/phoonnx_tts/sabela_cotovia_vits.json",
        "/home/miro/PycharmProjects/phoonnx_tts/celtia_vits.json",
        "/home/miro/PycharmProjects/phoonnx_tts/mimic3_gruut.json",
        "/home/miro/PycharmProjects/phoonnx_tts/mimic3_espeak.json",
        "/home/miro/PycharmProjects/phoonnx_tts/mimic3_epitran.json",
        "/home/miro/PycharmProjects/phoonnx_tts/mimic3_symbols.json",
        "/home/miro/PycharmProjects/phoonnx_tts/piper_espeak.json",
        "/home/miro/PycharmProjects/phoonnx_tts/vits-coqui-pt-cv/config.json",
        "/home/miro/PycharmProjects/phoonnx_tts/phonikud/model.config.json"
    ]
    phoneme_txts = [
        None,
        None,
        "/home/miro/PycharmProjects/phoonnx_tts/mimic3_ap/phonemes.txt",
        "/home/miro/PycharmProjects/phoonnx_tts/mimic3_ap/phonemes.txt",
        "/home/miro/PycharmProjects/phoonnx_tts/mimic3_ap/phonemes.txt",
        "/home/miro/PycharmProjects/phoonnx_tts/mimic3_ap/phonemes.txt",
        None,
        None,
        None
    ]
    print("Testing model config file parsing\n###############")
    for idx, cfile in enumerate(config_files):
        print(f"\nConfig file: {cfile}")
        with open(cfile) as f:
            config = json.load(f)
        print("Mimic3:", VoiceConfig.is_mimic3(config))
        print("Piper:", VoiceConfig.is_piper(config))
        print("Coqui:", VoiceConfig.is_coqui_vits(config))
        print("Phoonx:", VoiceConfig.is_phoonnx(config))
        cfg = VoiceConfig.from_dict(config, phoneme_txts[idx])
        print(cfg)

